#!/usr/bin/python3

from __future__ import annotations
from requests import get
from json import load
from typing import Dict, Any
from .model.url import RequestURL
from .model.data import Data
from .model.pollutant import Pollutant
from .model.station import Station


def _isStationAlreadyExisiting(name: str, data: Data) -> bool:
    return data.get(name) != None


'''
    Returns an instance of Station object, holding identifying
    information related to that Station.

    Pollutant objects ( i.e. air quality data collected ), 
    collected by this Station are to be stored under this Station object.
'''


def _makeStationObj(data: Dict[str, str]) -> Station:
    return Station(data['station'], data['city'], data['state'], data['country'], {})


'''
    Returns an instance of Pollutant object,
    holding data passed while invoking method.

    This instance is to be stored in resulting 
    Data object, under some Station object.
'''


def _makePollutantObj(data: Dict[str, str], timeStamp: int) -> Pollutant:
    return Pollutant(data['pollutant_id'], data['pollutant_min'], data['pollutant_max'], data['pollutant_avg'], data['pollutant_unit'], timeStamp, data['station'])


'''
    Parses fetched data and collects that into Data object,
    which is passed while invoking this method.
'''


def _parse(content: Dict[str, Any], data: Data):
    _timeStamp = content['updated']
    for v in content['records']:
        pollutantObj = _makePollutantObj(v, _timeStamp)
        if _isStationAlreadyExisiting(v['station'], data):
            data.updateStationRecord(pollutantObj)
        else:
            data.push(_makeStationObj(v)).updateStationRecord(pollutantObj)


'''
    Fetches data from specified URL.
'''


def _fetch(url: str) -> Dict[str, Any]:
    resp = get(url)
    content = resp.json() if resp.ok else None
    resp.close()
    return content


'''
    Collects air quality monitoring data from remote,
    and parsed data is put into instance of Data, which is
    passed while invoking method.

    Same instance gets modified and returned, after collection
    is completed.
'''


def request(_data: Data) -> Data:
    try:
        _req = RequestURL()
        for i in _req:
            _tmp = _fetch(i)
            if _tmp:
                _parse(_tmp, _data)
        return _data
    except Exception:
        return _data


'''
    If you've some data already collected by airQ,
    then passing that file path to this method,
    allows you to build an object to Data class,
    which is holding parsed data.

    Now you can simply modify this object
    i.e. send this Data object to `request()` method,
    written just above.

    Which will collect new data available, and when exported to JSON,
    will remove all data older than requested timestamp.
'''


def parse(source: str) -> Data:
    try:
        with open(source, 'r') as fd:
            return Data.fromJSON(load(fd))
    except Exception:
        return Data([])


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import Dict, Any

'''
    Holds record of a Pollutant, which are identified
    using one id.

    There're two very important fields `timeStamp` & `station`
'''


class Pollutant:
    def __init__(self, id: str, min: int, max: int, avg: int, unit: str, timestamp: int, station: str):
        self._id = id
        self._min = min
        self._max = max
        self._avg = avg
        self._unit = unit
        self._timestamp = timestamp
        self.forStation = station

    def toJSON(self) -> Dict[str, Any]:
        return {'id': self._id, 'min': self._min, 'max': self._max, 'avg': self._avg, 'unit': self._unit}

    @staticmethod
    def fromJSON(data: Dict[str, Any], timeStamp: int, station: str) -> Pollutant:
        return Pollutant(data['id'], data['min'], data['max'], data['avg'], data['unit'], timeStamp, station)


if __name__ == '__main__':
    print('[!]This module is designed to be work as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import Dict, List, Any, Tuple, Set
from .pollutant import Pollutant
from datetime import datetime, timedelta
from operator import sub


class Station:
    def __init__(self, name: str, city: str, state: str, country: str, records: Dict[int, List[Pollutant]]):
        self.name = name
        self.city = city
        self.state = state
        self.country = country
        self.records = records

    def _checkRedundancy(self, pollutantId: str, container: List[Pollutant]) -> bool:
        return False if len([i for i in container if i._id == pollutantId]) == 0 \
            else True

    def push(self, record: Pollutant) -> Station:
        _holder = self.records.get(record._timestamp)
        if not _holder:
            _holder = [record]
            self.records[record._timestamp] = _holder
        else:
            if not self._checkRedundancy(record._id, _holder):
                _holder.append(record)
        return self

    def toJSON(self) -> Dict[str, Any]:
        return {'name': self.name, 'city': self.city, 'state': self.state, 'country': self.country, 'records': dict([(k, [i.toJSON() for i in v]) for k, v in self.records.items()])}

    @staticmethod
    def fromJSON(data: Dict[str, Any]) -> Station:
        return Station(data['name'], data['city'], data['state'], data['country'], dict([(int(k), [Pollutant.fromJSON(i, int(k), data['name']) for i in v]) for k, v in data['records'].items()]))

    def removeOutOfRangeValues(self, _range: int):
        for i in [i for i in self.records.keys() if i < _range]:
            self.records.pop(i)
        return self

    '''
        Returns a tuple of datetime objects,
        where first one denotes oldest timestamp, at which we collected data,
        and second one denotes newest timestamp. We've pollutant data at this timestamp too.

        Difference of these two gives timespan, which is returned by method below ( self.getTimeSpan )
    '''
    @property
    def getTimeRange(self) -> Tuple[datetime, datetime]:
        _timestamps = list(self.records.keys())
        return datetime.fromtimestamp(min(_timestamps)), datetime.fromtimestamp(max(_timestamps))

    '''
        Returns calculated timedelta i.e. difference between
        highest timestamp and lowest timestamp
    '''
    @property
    def getTimeSpan(self) -> timedelta:
        def _swap(a, b):
            return b, a
        return sub(*_swap(*self.getTimeRange))

    '''
        How many # of records are present for this Station ?

        By record, I mean, how many number of times we collected data
        for this station ( possibly with 1 hour gap ), and that's present in this instance
    '''
    @property
    def recordCount(self) -> int:
        return len(self.records)

    '''
        Returns a set of all pollutant ids, for which we can lookup record

        Set returned, so that removal of duplicacy becomes easier.
    '''
    @property
    def getAvailablePollutants(self) -> Set[str]:
        return set([i._id for _, v in self.records.items() for i in v])

    '''
        Returns a list of Pollutant objects of a certain id.

        This method will be helpful, when plotting pollutant stat,
        for a certain pollutant type.
    '''

    def getPollutantStatByID(self, id: str) -> List[Pollutant]:
        return [i for _, v in self.records.items() for i in v if id == i._id]


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)



#!/usr/bin/python3

from __future__ import annotations
from typing import List, Dict, Any
from .station import Station
from .pollutant import Pollutant


class Data:
    def __init__(self, stations: List[Station]):
        self._stations = stations

    def __push__(self, name: str, low: int, high: int) -> int:
        if low > high:
            return 0
        elif low == high:
            return low if self._stations[low].name > name else (low + 1)
        else:
            mid = low + (high - low) // 2
            return self.__push__(name, low, mid) \
                if self._stations[mid].name > name \
                else self.__push__(name, mid + 1, high)

    '''
        Insert a new Station object into Data object,
        while keeping sorted order intact.

        Well sorting is done
        using air quality monitoring station's names ( ascendingly )
    '''

    def push(self, record: Station) -> Data:
        self._stations.insert(self.__push__(
            record.name, 0, len(self._stations) - 1), record)
        return self

    def _get(self, name: str, low: int, high: int) -> int:
        if low > high:
            return -1
        elif low == high:
            return low if self._stations[low].name == name else -1
        else:
            mid = low + (high - low) // 2
            return self._get(name, low, mid) \
                if self._stations[mid].name >= name \
                else self._get(name, mid + 1, high)

    '''
        Given a air quality monitoring station's name,
        it finds out Station instance with that name,
        using binary search ( well of course from a sorted array )
    '''

    def get(self, name: str) -> Station:
        _tmp = self._get(name, 0, len(self._stations) - 1)
        return None if _tmp == -1 else self._stations[_tmp]

    def updateStationRecord(self, record: Pollutant):
        self.get(record.forStation).push(record)
        return self

    '''
        Converts to JSON, to be invoked
        when we need to export collected data into JSON,
        for future reference
    '''

    def toJSON(self, _range: int) -> Dict[str, Any]:
        return {'stations': [i.toJSON() for i in self._removeOutOfRangeValues(_range)._stations]}

    '''
        Parse JSON data, and return Data object, holding that
        dataset ( which was previously exported into JSON )
    '''
    @staticmethod
    def fromJSON(data: Dict[str, Any]) -> Data:
        return Data([Station.fromJSON(i) for i in data['stations']])

    '''
        Returns max time stamp ( i.e. most recent time stamp
        for which we've data ), which will be helpful in eliminating
        those records which are having time stamp lesser than limit
        set ( i.e. older than a certain time )
    '''
    @property
    def _getMaxTimeStamp(self) -> int:
        return max([i for i in self._stations[0].records.keys()])

    '''
        Remove those pollutant records which are having
        their corresponding timestamp lesser than time range set currently.
    '''

    def _removeOutOfRangeValues(self, _range: int):
        _tmp = self._getMaxTimeStamp - _range
        for i in self._stations:
            i.removeOutOfRangeValues(_tmp)
        return self

    '''
        Returns # of air quality monitoring station count,
        which is present in current dataset
    '''
    @property
    def getStationCount(self):
        return len(self._stations)

    '''
        Returns a dictionary, where keys are state names ( spread across India ),
        and corresponding values denote air quality monitoring station names,
        which are belonging to that city
    '''
    @property
    def getStationsGroupedByState(self) -> Dict[str, List[str]]:
        _grouped = {}
        for i in self._stations:
            _holder = _grouped.get(i.state)
            if _holder:
                _holder.append(i.name)
            else:
                _grouped[i.state] = [i.name]
        return _grouped

    '''
        Returns a dictionary, where keys are city names ( spread across India ),
        and corresponding values denote air quality monitoring station names,
        which are belonging to that city
    '''
    @property
    def getStationsGroupedByCity(self) -> Dict[str, List[str]]:
        _grouped = {}
        for i in self._stations:
            _holder = _grouped.get(i.city)
            if _holder:
                _holder.append(i.name)
            else:
                _grouped[i.city] = [i.name]
        return _grouped

    '''
        Returns a dictionary holding states as its key values,
        and a list of city names which are part of that State,
        from which we've air quality monitoring data
    '''
    @property
    def getCitiesUnderStates(self) -> Dict[str, List[str]]:
        return dict([(k, set([self.get(i).city for i in v])) for k, v in self.getStationsGroupedByState.items()])

    '''
        Returns a list of air quality monitoring station names
    '''
    @property
    def getStationNames(self) -> List[str]:
        return [i.name for i in self._stations]


if __name__ == '__main__':
    print('[!]This module is designed to be used as a backend handler')
    exit(0)
